unit DownloadThread;

interface
uses Windows, Classes, Graphics, SysUtils, IdHttp, RVFuncs;

type
  TImageDownloadThread = class (TThread)
    private
      FIdHttp: TIdHTTP;
      FURL: String;
      FDefWidth, FDefHeight: Integer;
      FGraphic: TGraphic;
      FTempFileName: String;
      procedure MakeGraphic;
    public
      constructor Create(const URL: String; DefWidth, DefHeight: Integer);
      destructor Destroy; override;
      procedure Execute; override;
      property Graphic: TGraphic read FGraphic;
      property URL: String read FURL;
  end;

function CreatePlaceHolderImage(Width, Height: Integer): TBitmap;
function CreateErrorImage(Width, Height: Integer): TBitmap;

implementation

{------------------------------------------------------------------------------}
{ Creates image to display while downloading }
function CreatePlaceHolderImage(Width, Height: Integer): TBitmap;
begin
  if Width<=0 then
    Width := 50;
  if Height<=0 then
    Height := 50;
  Result := TBitmap.Create;
  Result.Width := Width;
  Result.Height := Height;
  Result.Canvas.Pen.Color := clGray;
  Result.Canvas.Rectangle(0, 0, Width, Height);
  Result.Canvas.Font.Name := 'Arial';
  Result.Canvas.TextOut(5, 5, 'Downloading...');
end;
{------------------------------------------------------------------------------}
{ Creates image to display in case of error }
function CreateErrorImage(Width, Height: Integer): TBitmap;
begin
  if Width<=0 then
    Width := 100;
  if Height<=0 then
    Height := 10;
  Result := TBitmap.Create;
  Result.Width := Width;
  Result.Height := Height;
  Result.Canvas.Pen.Color := clGray;
  Result.Canvas.Rectangle(0, 0, Width, Height);
  Result.Canvas.Font.Name := 'Arial';
  Result.Canvas.Font.Color := clRed;
  Result.Canvas.TextOut(5, 5, '?!');
end;
{------------------------------------------------------------------------------}
{ Returns the system directory for temporal files }
function GetTempDir: String;
var l: Integer;
begin
  // Warning: : The GetTempPath function
  // does not verify that the returned
  // directory exists.
  SetLength(Result, 300);
  l := GetTempPath(300, PChar(Result));
  SetLength(Result, l);
end;
{------------------------------------------------------------------------------}
{ Returns a name for the temporal file with the given extension }
function GetTempFileName(const Ext: String): String;
var Path: String;
    v: Integer;
begin
  Path := GetTempDir+'tmp';
  v := Random(MaxInt div 2);
  repeat
    inc(v);
    Result := Path+IntToStr(v)+Ext;
  until not FileExists(Result);
end;
{------------------------------------------------------------------------------}
function StripFileName(const FileName: String): String;
var p: Integer;
begin
  p := Pos('?', FileName);
  if p<>0 then
    Result := Copy(FileName, 1, p-1)
  else
    Result := FileName;
end;
{============================== TImageDownloadThread ==========================}
constructor TImageDownloadThread.Create(const URL: String; DefWidth, DefHeight: Integer);
begin
  inherited Create(True);
  FIdHttp := TIdHTTP.Create(nil);
  FURL := URL;
  FDefWidth := DefWidth;
  FDefHeight := DefHeight;
end;
{------------------------------------------------------------------------------}
destructor TImageDownloadThread.Destroy;
begin
  FreeAndNil(FGraphic);
  FreeAndNil(FIdHttp);
  inherited;
end;
{------------------------------------------------------------------------------}
procedure TImageDownloadThread.Execute;
var FileStream: TFileStream;
begin
  FTempFileName := GetTempFileName(ExtractFileExt(StripFileName(FURL)));
  try
    FileStream := TFileStream.Create(FTempFileName, fmCreate);
    try
      FIdHTTP.Get(FURL, FileStream);
    finally
      FileStream.Free;
    end;
  except
    FTempFileName := '';
  end;
  Synchronize(MakeGraphic);
end;
{------------------------------------------------------------------------------}
procedure TImageDownloadThread.MakeGraphic;
var pic: TPicture;
begin
  if FTempFileName='' then begin
    FGraphic := CreateErrorImage(FDefWidth, FDefHeight);
    exit;
  end;
  pic := TPicture.Create;
  try
    pic.LoadFromFile(FTempFileName);
    FGraphic := RV_CreateGraphics(TGraphicClass(pic.Graphic.ClassType));
    FGraphic.Assign(pic.Graphic);
  finally
    pic.Free;
    DeleteFile(FTempFileName);
  end;
end;

end.
