{*******************************************************}
{                                                       }
{       RichView                                        }
{       Demo: using TRVHtmlImporter and Indy to         }
{       download HTML files with pictures.              }
{       Pictures are downloaded in threads.             }
{                                                       }
{       Copyright (c) Sergey Tkachenko                  }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}

{==============================================================================
  The main things that must be improved in this demo:
  1. Limiting number of downloading threads. If there are too many threads,
     new threads must wait until old threads complete their work.
  2. Interrupting. In this demo, you cannot open a new address
     or close the application while downloading is in progress.
==============================================================================}

unit rvhtmltest1;

interface

{$DEFINE USE_GIFIMAGE}
{$DEFINE USE_PNGOBJECT}

uses
  // Borland's
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, ShellApi, ComCtrls,
  // TRichView
  RVScroll, RichView, RVStyle, CRVData, CRVFData, RVFuncs, RVNormalize, RVItem,
  RVTable,
  rvhtmlimport,
  // Graphics
  {$IFDEF USE_GIFIMAGE}
  GifImage,
  {$ENDIF}
  {$IFDEF USE_PNGOBJECT}
  PngImage,
  {$ENDIF}
  // Indy
  IdBaseComponent, IdComponent, IdTCPConnection, IdTCPClient, IdHTTP,
  DownloadThread;

{
  Remove dots from the defines below to enable support for Gif and Png.

  GifImage (by Anders Melander)
    http://www.torry.net/vcl/graphics/gif/gifimage.exe (original)
    http://www.trichview.com/resources/thirdparty/gifimage.zip (update)
  PngObject (by Gustavo Huffenbacher Daud)
    http://pngdelphi.sourceforge.net/
}




type
  TForm1 = class(TForm)
    RVStyle1: TRVStyle;
    RichView1: TRichView;
    RvHtmlImporter1: TRvHtmlImporter;
    Panel1: TPanel;
    btnOpen: TButton;
    OpenDialog1: TOpenDialog;
    IdHTTP1: TIdHTTP;
    btnPaste: TButton;
    StatusBar1: TStatusBar;
    Button1: TButton;
    Edit1: TEdit;
    procedure RvHtmlImporter1ImageRequired2(Sender: TObject;
      const src: String; Width, Height: Integer; var Img: TGraphic);
    procedure RichView1Jump(Sender: TObject; id: Integer);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
  private
    { Private declarations }
    DownloadedImages: TStringList;
    DownloadThreads: TThreadList;
    Formatted: Boolean;
    function GetDownloadingImagesCount: Integer;
    procedure ClearCache;
    procedure DoThreadTerminate(Sender: TObject);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}
{------------------------------------------------------------------------------}
procedure TForm1.FormCreate(Sender: TObject);
begin
  // Creating memory cache for downloaded images
  DownloadedImages := TStringList.Create;
  DownloadedImages.Sorted := True;
  DownloadedImages.Duplicates := dupIgnore;
  DownloadedImages.CaseSensitive := False;
  // Creating list of images downloading threads
  DownloadThreads := TThreadList.Create;
end;
{------------------------------------------------------------------------------}
procedure TForm1.FormDestroy(Sender: TObject);
begin
  DownloadThreads.Free;

  ClearCache;
  DownloadedImages.Free;
end;
{------------------------------------------------------------------------------}
{ ImageRequired2 event: providing image }
procedure TForm1.RvHtmlImporter1ImageRequired2(Sender: TObject;
  const src: String; Width, Height: Integer; var Img: TGraphic);
var pic: TPicture;
    Img2: TGraphic;
    Index: Integer;
    FileName: String;
    Thread: TImageDownloadThread;
begin
  try
    Img := nil;
    if Pos(':', src)>0 then
      FileName := src  // absolute path
    else
      FileName := RvHtmlImporter1.BasePath+src; // relative path
    // If this image was already downloaded (or is being downloaded), taking
    // it from cache
    if DownloadedImages.Find(FileName, Index) then begin
      Img := RV_CreateGraphics(TGraphicClass(TGraphic(DownloadedImages.Objects[Index]).ClassType));
      Img.Assign(TGraphic(DownloadedImages.Objects[Index]));
      exit;
    end;
    // If not, downloading or reading locally
    if Pos('http://', LowerCase(FileName))=1 then begin
      { DOWNLOADING }
      Img := CreatePlaceHolderImage(Width, Height);
      Thread := TImageDownloadThread.Create(FileName, Width, Height);
      Thread.FreeOnTerminate := True;
      DownloadThreads.Add(Thread);
      Thread.OnTerminate := DoThreadTerminate;
      StatusBar1.SimpleText := 'Downloading '+IntToStr(GetDownloadingImagesCount)+' images...';
      Thread.Resume;
      end
    else begin
      { READING LOCAL FILE }
      pic := TPicture.Create;
      try
        pic.LoadFromFile(FileName);
        Img := RV_CreateGraphics(TGraphicClass(pic.Graphic.ClassType));
        Img.Assign(pic.Graphic);
      finally
        pic.Free;
      end;
    end;
    if Img<>nil then begin
      // Adding image in cache. This image can be either read locally, or
      // queued for downloading
      Img2 := RV_CreateGraphics(TGraphicClass(Img.ClassType));
      Img2.Assign(Img);
      DownloadedImages.AddObject(FileName, Img2)
    end;
  except
    Img.Free;
    Img := nil;
  end;
  if Img=nil then begin
    { LOCAL READING FAILED. CREATING DEFAULT IMAGE }
    Img := CreateErrorImage(Width, Height);
    Img2 := RV_CreateGraphics(TGraphicClass(Img.ClassType));
    Img2.Assign(Img);
    DownloadedImages.AddObject(FileName, Img2)
  end;
  StatusBar1.SimpleText := '';
end;
{------------------------------------------------------------------------------}
{ OnJump occurs when hyperlink is clicked }
procedure TForm1.RichView1Jump(Sender: TObject; id: Integer);
var
  RVData: TCustomRVFormattedData;
  ItemNo: Longint;
begin
  RichView1.GetJumpPointLocation(id, RVData, ItemNo);
  //ShellExecute(0, 'open', PChar(RVData.GetItemTag(ItemNo)), nil, nil, SW_SHOW);
  //Application.MessageBox(PChar(RVData.GetItemTag(ItemNo)), 'Clicked',
  //  MB_OK or MB_ICONINFORMATION)
  Edit1.Text := PChar(RVData.GetItemTag(ItemNo));
  Edit1.SetFocus;
end;
{------------------------------------------------------------------------------}
{ Clearing cache of downloaded images }
procedure TForm1.ClearCache;
var i: Integer;
begin
  for i := 0 to DownloadedImages.Count-1 do
    DownloadedImages.Objects[i].Free;
  DownloadedImages.Clear;
end;
{------------------------------------------------------------------------------}
{ Replaces all placeholders for the picture downloaded from URL.
  Placeholders are changed to NewImg }
procedure ChangePicture(RVData: TCustomRVData; const URL: String; NewImg: TGraphic);
var i, r, c, Tag: Integer;
  VAlign: TRVVAlign;
  ItemName, ItemURL: String;
  Table: TRVTableItemInfo;
  Img, Img2: TGraphic;
begin
  for i := 0 to RVData.ItemCount-1 do
    case RVData.GetItemStyle(i) of
      rvsPicture, rvsHotPicture:
        begin
          RVData.GetItemExtraStrProperty(i, rvespImageFileName, ItemURL);
          if CompareText(ItemURL, URL)=0 then begin
            RVData.GetPictureInfo(i, ItemName, Img, VAlign, Tag);
            Img2 := RV_CreateGraphics(TGraphicClass(NewImg.ClassType));
            Img2.Assign(NewImg);
            RVData.SetPictureInfo(i, ItemName, Img2, VAlign, Tag);
          end;
        end;
      rvsTable:
        begin
          Table := TRVTableItemInfo(RVData.GetItem(i));
          RVData.GetItemExtraStrProperty(i, rvespImageFileName, ItemURL);
          if CompareText(ItemURL, URL)=0 then
            Table.BackgroundImage := NewImg;
          for r := 0 to Table.RowCount-1 do
            for c := 0 to Table.ColCount-1 do
              if Table.Cells[r,c]<>nil then begin
                if CompareText(Table.Cells[r,c].BackgroundImageFileName, URL)=0 then
                  Table.Cells[r,c].BackgroundImage := NewImg;
                ChangePicture(Table.Cells[r,c].GetRVData, URL, NewImg);
              end;
        end;
    end;
end;
{------------------------------------------------------------------------------}
{ In OnTerminate for image downloading thread, we do the following:
  1) changing all placeholders for Thread.URL to copies of Thread.Graphic
  2) updating image in memory cache (DownloadedImages)
  3) removing thread from list (DownloadThreads)
  4) reformatting RichView1, if it was formatted. }
procedure TForm1.DoThreadTerminate(Sender: TObject);
var Thread: TImageDownloadThread;
    Img: TGraphic;
    Index, Count: Integer;
begin
  Thread := TImageDownloadThread(Sender);
  ChangePicture(RichView1.RVData, Thread.URL, Thread.Graphic);
  Img := RV_CreateGraphics(TGraphicClass(Thread.Graphic.ClassType));
  Img.Assign(Thread.Graphic);
  if DownloadedImages.Find(Thread.URL, Index) then begin
    DownloadedImages.Objects[Index].Free;
    DownloadedImages.Objects[Index] := Img;
  end;
  DownloadThreads.Remove(Thread);
  Count := GetDownloadingImagesCount;
  if Count=0 then
    StatusBar1.SimpleText := ''
  else
    StatusBar1.SimpleText := 'Downloading '+IntToStr(GetDownloadingImagesCount)+' images...';
  if Formatted then
    RichView1.Reformat;
end;
{------------------------------------------------------------------------------}
{ Loading HTML by the address in Edit1.Text.
  IMPORTANT: if index file (such is index.html) is omitted, Edit1.Text must be
  finished with '/', otherwise RvHtmlImporter1.BasePath will not be assigned
  properly }
procedure TForm1.Button1Click(Sender: TObject);
var s: String;
begin
  if GetDownloadingImagesCount>0 then begin
    Application.MessageBox('Please wait while downloading is in progress', 'Cannot open new page', MB_OK);
    exit;
  end;

  if Edit1.Text='' then
    exit;
  Screen.Cursor := crHourGlass;
  try
    Formatted := False;
    ClearCache;
    IdHTTP1.HandleRedirects := True;
    s := IdHTTP1.Get(Edit1.Text);
    StatusBar1.SimpleText := '';
    if Edit1.Text[Length(Edit1.Text)]='/' then
      RvHtmlImporter1.BasePath := Edit1.Text
    else
      RvHtmlImporter1.BasePath := ExtractFilePath(Edit1.Text);
    RvHtmlImporter1.LoadHtml(s);
    NormalizeRichView(RichView1.RVData);
  finally
    RichView1.Format;
    Formatted := True;
    Screen.Cursor := crDefault;
  end;
end;
{------------------------------------------------------------------------------}
{ Returns number of active image downloading threads }
function TForm1.GetDownloadingImagesCount: Integer;
begin
  Result := DownloadThreads.LockList.Count;
  DownloadThreads.UnlockList;
end;
{------------------------------------------------------------------------------}
{ Cannot close while downloading }
procedure TForm1.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
 CanClose := GetDownloadingImagesCount=0;
 if not CanClose then
   Application.MessageBox('Please wait while downloading is in progress', 'Cannot open new page', MB_OK);
end;

initialization
{$IFDEF USE_GIFIMAGE}
  TPicture.RegisterFileFormat('gif', 'Gif Image', TGifImage);
  RegisterClasses([TGifImage]);
{$ENDIF}
{$IFDEF USE_PNGOBJECT}
  TPicture.RegisterFileFormat('png', 'Png Image', TPngObject);
  RegisterClass(TPngObject);
{$ENDIF}

end.
